Require Import Ascii.
Require Import List.
Require Import Nat.
Require Import String.
Require Import UntypedLC.

Include LC.

Set Implicit Arguments.


(*
  Notación auxiliar para manipulación de listas.
*)

Notation "x :: l" := (cons x l) (at level 60, right associativity).
Notation "[ ]" := nil.
Notation "[ x ; .. ; y ]" := (cons x .. (cons y nil) ..).

(*
  Implementación auxiliar de Coq.Init.Nat.ltb
  para facilitar pruebas posteriores.
*)

Fixpoint bltb (a b : nat) : bool :=
  match a, b with
  | O, S x => true
  | S x, S y => bltb x y
  | _, _ => false
  end.

(* 
  Función que busca por una cadena `s' en un término de LTerm `t';
  regresa `true' si esta existe.
*)

Fixpoint search_st (s: string) (t: LTerm) : bool :=
  match t with
  | LVar x => if string_dec x s then true else false
  | LAbs a b => if string_dec s a then true else (search_st s b)
  | LApp a b => orb (search_st s a) (search_st s b)
  end.

(*
  Función que busca por una cadena `s' en una lista 'l' de cadenas;
  regresa `true' si esta existe.
*)

Fixpoint bsearch_sl (s: string) (l: list (string)) : bool :=
  match l with
  | nil => false
  | a :: b =>
    if string_dec s a then true
    else bsearch_sl s b
  end.

(*
  Función que busca por una cadena `s' en una lista de tuplas (string * nat) `l';
  regresa el número asociado a esta si ella existe.
*)

Fixpoint search_snl (s: string) (l : list (string * nat)) : option nat :=
  match l with
  | nil => None
  | (a, b) :: t => if string_dec s a then Some b else search_snl s t
  end.

(*
  Función que remueve duplicados de una lista.
*)

Fixpoint rem_dup_list_str' (l k: list (string)) : list (string) :=
  match l with
  | a :: b =>
    if bsearch_sl a k then rem_dup_list_str' b k
    else rem_dup_list_str' b (a :: k)
  | nil => k
  end.

Definition rem_dup_list_str (l : list string) : list string :=
  rem_dup_list_str' l nil.

(*
  Función que, dado un término LTerm, obtiene la lista de todas sus variables.
*)

Fixpoint get_sl' (t: LTerm) (l: list string) : list string :=
  match t with
  | LVar x => if bsearch_sl x l then l else x :: l
  | LAbs a b => if bsearch_sl a l then get_sl' b l else a :: (get_sl' b l)
  | LApp a b => (get_sl' a l) ++ (get_sl' b l)
  end.

Definition get_sl (t: LTerm) : list string :=
  rem_dup_list_str (get_sl' t nil).

(*
  Dada una lista `l' de tuplas (string * natural), agrega uno a cada
  número natural de la lista.
*)

Fixpoint add_one' (l k : list (string * nat)) : list (string * nat) :=
  match l with
  | (x, n) :: t => add_one' t ((x, n+1) :: k)
  | nil => k
  end.

Definition add_one (l : list (string * nat)) : list (string * nat) :=
  add_one' l nil.

(* 
  Dada una lista `l' de tuplas (string * natural), una cadena `s' y
  un número natural `n', busca por `s' dentro de `l' y substrae `n' de
  su tupla.
*)

Fixpoint sub_snl' (l k : list (string * nat)) (s : string) (n : nat)
  : list (string * nat) :=
  match l with
  | (x, y) :: t =>
    if string_dec s x
    then sub_snl' t ((x, y - n) :: k) s n
    else sub_snl' t ((x, y) :: k) s n
  | nil => k
  end.

Definition sub_snl (l : list (string * nat)) (s : string) (n : nat)
  : list (string * nat) := sub_snl' l nil s n.

(*
  Función que convierte una cadena a una lista de caracteres ASCii.
*)

Fixpoint string_al' (s : string) (l : list ascii) : list ascii :=
  match s with
    | EmptyString => l
    | String c rest => string_al' rest (c :: l)
  end.

Definition string_al (s: string) : list ascii :=
  rev (string_al' s nil).

(*
  Función que convierte un valor booleano a un número natural, 0 ó 1.
*)

Definition nat_of_bool (b : bool) := if b then 1 else 0.

Coercion nat_of_bool : bool >-> nat.

(*
  Función que convierte a un caracter ASCii a un número natural:

  a => 1
  b => 2
  ...
  z => 26
*)

Definition nat_of_ascii_char a :=
  match a with
   Ascii b1 b2 b3 b4 b5 b6 b7 b8 =>
     b1 + 2 * (b2 + 2 * (b3 + 2 * (b4 + 2 *
      (b5 + 2 * (b6 + 2 * (b7 + 2 * b8)))))) - 96
  end.

(*
  Bien ordenamos las cadenas, de forma que podamos
  asignar de manera correcta índices de de Bruijn a
  variables libres:

  a => 1
  b => 2
  ...
  z => 26
  aa => 27
  ab => 28
  ...
  az => 52
  ...
*)

Fixpoint context_value' (l : list ascii) (n c : nat) : nat :=
  match l with
  | h :: t => context_value' t (n + (pow 26 c) * (nat_of_ascii_char h)) (c-1)
  | nil => n + 100 (* Free variables start at 101 *)
  end.

Definition context_value (s : string) : nat :=
  context_value' (string_al s) 0 (Datatypes.length (string_al s) - 1).

Fixpoint sl_context_snl' (l : list string) (k : list (string * nat)) :
  list (string * nat) :=
  match l with
  | h :: t => sl_context_snl' t ((h, context_value h) :: k)
  | nil => k
  end.

Definition sl_context_snl (l : list string) : list (string * nat) :=
  sl_context_snl' l nil.

(*
  Función que, dada una lista `l' de tuplas (string * nat),
  y una cadena `s', busca por `s' dentro de `l' y regresa su
  valor numérico asociado.
*)

Fixpoint getn_snl (s : string) (l : list (string * nat)) : option nat :=
  match l with
  | (h, n) :: t => if string_dec s h then Some n else getn_snl s t
  | nil => None
  end.

(*
  Función que busca por un término t en una lista 'l' de términos;
  regresa `true' si este existe.
*)

Fixpoint bsearch_tl (t: Term) (l: list Term) : bool :=
  match l with
  | nil => false
  | a :: b =>
    if eqterm t a then true
    else bsearch_tl t b
  end.

(*
  Dada una lista `l` de términos del cálculo lambda no tipado
  usando índices de de Bruijn, devolver la misma sin duplicados.
*)

Fixpoint rem_dup_list_term' (l k: list Term) : list Term :=
  match l with
  | a :: b =>
    if bsearch_tl a k then rem_dup_list_term' b k
    else rem_dup_list_term' b (a :: k)
  | nil => k
  end.

Definition rem_dup_list_term (l : list Term) : list LC.Term :=
  rem_dup_list_term' l nil.

Definition D := Abs ((Var 1) @ (Var 1)).
Definition B := Abs (Abs (Abs ((Var 3) @ ((Var 2) @ (Var 1))))).
Definition W (F : Term) := B @ F @ D.
Definition X (F : Term) := D @ (W F).
Definition Y := Abs ((Abs ((Var 2) @ ((Var 1) @ (Var 1))))
                @ (Abs ((Var 2) @ ((Var 1) @ (Var 1))))).

Fixpoint lnat_to_nat (n : Term) : nat :=
  match n with
  | Abs (Abs (Var 1)) => 0
  | lnat_succ @ m => S (lnat_to_nat m)
  | _ => 0
  end.

(*
  Función para pasar de términos del cálculo
  lambda no tipado común LTerm a términos utilizando
  índices de de Bruijn.
*)

Fixpoint lterm_to_term' (t : LTerm) (l : list (string * nat)) : Term :=
  match t with
  | LVar x =>
    match getn_snl x l with
    | Some n => Var n
    | None => Var (context_value x)
    end
  | LAbs a b => Abs (lterm_to_term' b (add_one((a,0)::l)))
  | LApp a b => App (lterm_to_term' a l) (lterm_to_term' b l)
  end.

Definition lterm_to_term (t : LTerm) : Term :=
  lterm_to_term' t  nil.

(*
  Función que obtiene una lista con todos los
  lambda términos que son variables libres de una
  lambda expresión t.
*)

Fixpoint free_var' (t : Term) (k : nat) : (list Term) :=
  match t with
  | Var i =>
    if bltb i k then [ Var i ]
    else nil
  | u @ v => free_var' u k ++ free_var' v k
  | Abs x => free_var' x (k + 1)
  end.

Definition free_var t := rem_dup_list_term (free_var' t 1).


Lemma rev_extensional_inference (A B : Prop) (f g : A -> B) (a : A) :
  (f = g) -> (f a = g a).
Proof.
  intro.
  rewrite H.
  reflexivity.
Qed.

Lemma decidable_nats : forall (M N : nat), M = N \/ M <> N.
Proof.
  double induction M N.
  left.
  reflexivity.
  auto.
  auto.
  intros.
  elim (H0 n).
  auto.
  auto.
Qed.

Lemma decidable_terms : forall (M N : Term), M = N \/ M <> N.
Proof.
  induction M;
  induction N;
  repeat try (left; congruence);
  repeat try (right; congruence).
  elim (decidable_nats n0 n);
  auto.
  intros.
  right.
  congruence.
  elim (IHM N).
  intros.
  left.
  congruence.
  intros.
  right.
  congruence.
  elim(IHM1 N1).
  elim(IHM2 N2).
  left.
  congruence.
  right.
  congruence.
  right.
  congruence.
Qed.

Lemma bltb_n_0 : forall (n : nat), bltb n 0 = false.
Proof.
  intros.
  induction n.
  - reflexivity.
  - reflexivity.
Qed.

Lemma bltb_n_n : forall (n : nat), bltb n n = false.
Proof.
  intros.
  induction n.
  - reflexivity.
  - simpl.
    rewrite IHn.
    reflexivity.
Qed.

Lemma eqb_refl : forall x, PeanoNat.Nat.eqb x x = true.
Proof.
  intros.
  induction x.
  reflexivity.
  simpl.
  exact IHx.
Qed.