Require Import EqNat.
Require Import General_relations.
Require Import UntypedLC.
Require Import UntypedLC_aux.

Include LC.
Include GRels.

(*
  Operador de lifting para la substitución de términos
  a partir de índices de de Bruijn.
  Únicamente es relevante cuando se tienen variables
  libres, las cuales denotamos a partir de un índice
  establecido de forma arbiraria en el contexto usado.
*)

Fixpoint liftn (n : nat) (t : Term) (k : nat) : Term :=
  match t with
  | Var i => if bltb i k then Var i else Var (i + n)
  | App u v => App (liftn n u k) (liftn n v k)
  | Abs s => Abs (liftn n s (k + 1))
  end.

Definition lift (n: nat) (t: Term) := liftn n t 1.

(*
  Operación de substitución utilizada para la
  beta reducción.
*)

Fixpoint substn (t s : Term) (k : nat) : Term :=
  match t with
  | Var i =>
    if bltb k i then Var (i - 1)
    else if beq_nat i k then s
    else Var i
  | App u v => App (substn u s k) (substn v s k)
  | Abs u => Abs (substn u s (k + 1))
  end.

Definition subst (t s : Term) : Term := substn t s 1.

(*
Axiom beta : forall M N, ((Abs M) @ N) = subst M N.
Axiom eta : forall M, ((Abs M) @ (Var 1)) = M.

Hint Resolve beta eta : base0.
*)

(*
  Reglas de reducción.
*)

Inductive red : Term -> Term -> Prop :=
  | bred : forall M N, red ((Abs M) @ N) (subst M N)
  | appl : forall M N X, red M N -> red (M @ X) (N @ X)
  | appr : forall M N X, red M N -> red (X @ M) (X @ N)
  | abs : forall M N, red M N -> red (Abs M) (Abs N).

(*
  Reducciones múltiples ->*
*)

Definition reds := star Term red.

(*
  Cerradura transitiva de la reddución.
 *)

Definition redt := star_trans Term red.

Lemma substn_eq : forall (t : Term) (k : nat), substn (Var k) t k = t.
Proof.
  intros.
  induction k.
  simpl.
  reflexivity.
  unfold subst.
  simpl.
  rewrite bltb_n_n.
  rewrite eqb_refl.
  reflexivity.
Qed.

Lemma subst_gt : forall (t : Term) (i j : nat), 
  bltb i j = true -> substn (Var j) t i = Var (j - 1).
Proof.
  intros.
  simpl.
  rewrite -> H.
  reflexivity.
Qed.

Lemma appl_reds : forall M N X, reds M N -> reds (M @ X) (N @ X).
Proof.
  intros.
  induction H.
  apply star_refl.
  apply (star_R _ _ _ (y @ X) _).
  exact IHstar.
  apply appl.
  exact H0.
Qed.

Lemma appr_reds : forall M N X, reds M N -> reds (X @ M) (X @ N).
Proof.
  intros.
  induction H.
  apply star_refl.
  apply (star_R _ _ _ (X @ y) _).
  exact IHstar.
  apply appr.
  exact H0.
Qed.

Lemma abs_reds : forall M N, reds M N -> reds (Abs M) (Abs N).
Proof.
  intros.
  induction H.
  apply star_refl.
  apply (star_R _ _ _ (Abs y) _).
  apply IHstar.
  apply abs.
  exact H0.
Qed.

Lemma breds : forall M M' N N', reds M M' -> reds N N'
  -> reds ((Abs M') @ N') (subst M' N').
Proof.
  intros.
  induction H.
  apply (star_R _ _ _ (Abs x @ N') _).
  apply star_refl.
  apply bred.
  apply (star_R _ _ _ (Abs z @ N') _).
  apply star_refl.
  apply bred.
Qed.

Lemma d_behaviour : forall x, reds (D @ x) (x @ x).
Proof.
  intros.
  unfold D.
  apply (star_R _ _ _ (Abs (Var 1 @ Var 1) @ x) _).
  apply star_refl.
  apply bred.
Qed.